/*

'FLASH' 'S'tandard mode, designed for 'H'itech c (FLASH_SH)

this code accesses flash directly, for testing purposes

Processor:				pic17c43
Compiler:					hi-tech C v7.83pl3
Environment:			compiled with MPLab 4.00.14
Description:			flash routines to communicate with AM29F400 flash
Project setup:		must be in extended microprocessor mode, using ram supplied from target board
									linker options: -L-Pwfh=01FFEh -L-Pesh=01FFCh -L-Pfdih=01FFAh -L-Pefh=01FF8h
Optimisation:			this code is not optimised for speed, it needs some routines converted to
									#defines and the total drop path of procedures shortened
Author:						Shane Tolmie
Tested:						as part of system to download and run code from flash for remote web updates
									data acquisition portion tested OK
									
*/

#define bootrom_write_flash_hook 0x1FFE/2
#define bootrom_erase_sector_hook 0x1FFC/2
#define bootrom_flash_device_id_hook 0x1FFA/2		//rom is in words, .hex file in bytes
#define API_EraseFlash_Hook 0x1FF8/2

#include	<pic.h>
#include	"flash_sh.h"
#include	"codes.h"
#include	"always.h"
#include	"delay.h"
#include	"serial.h"

//variables for communication between the program running in flash, and that running in eeprom
bank1 unsigned int	api_flash_uint_control			@ 0x1FE;
bank1 unsigned int	api_flash_uint_address			@	0x1FC;
bank1 unsigned int	api_flash_uint_data					@	0x1FA;
bank1 unsigned char	api_flash_uint_hi_data			@ (0x1FA+1);
bank1 unsigned char	api_flash_uint_lo_data			@ (0x1FA+0);
bank1 unsigned int	api_flash_uint_working 			@ 0x1F8;
bank1 unsigned char	api_flash_uint_hi_working		@ (0x1F8+1);
bank1 unsigned char	api_flash_uint_lo_working		@ (0x1F8+0);
bank1 unsigned char	api_flash_uchar_working 		@ 0x1F7;
bank1 unsigned char	api_flash_uchar_return 			@ 0x1F6;
bank1 unsigned int	api_flash_rwrom_temp 				@ 0x1F4;

void API_SelectFlashBank(unsigned char bank)
{
	FLASH_A16=bank & 0x01;
}

unsigned int write_read_rom_data;

//uses table write to read data from external address lines
//the NOP instructions must be present for reliable operation

//NOTE: this routine has highly picky timing requirements,
//if it doesn't have the exact number of instructions shown
//here, it doesn't work, taking out a nop or adding one 
//makes it cease to perform correctly
//I don't know what will happen if the clock speed is changed!
unsigned int readrom(unsigned int address)
{
	TBLPTRH = (address >> 8);
	TBLPTRL = (address & 0xFF);
	#asm
	MOVLR 0x1																		//switch to bank1
	TABLRD 0,1,_api_flash_rwrom_temp						//dummy read to get it into sync
	TLRD 0,_api_flash_rwrom_temp
	NOP
	TABLRD 1,1,_api_flash_rwrom_temp+1
	NOP
	#endasm
	return api_flash_rwrom_temp;
}

unsigned int API_FlashDeviceID(void)
{
	void (*bootrom_flash_device_id_pointer)(void);							//declare pointer to function
	bootrom_flash_device_id_pointer
		=(void (*)(void))ReadFlash(bootrom_flash_device_id_hook);	//hook to vector, see top of this file
	bootrom_flash_device_id_pointer();													//equivalent to API_FlashDeviceID();
	return api_flash_uint_data;
}

//pins A16 and A17 are connected to RA3 and RB5
void API_SetupFlash(void)
{
	unsigned char dummy;

	//read RA3 to set it to an input
	#define FLASH_RY_BYBAR	RA3
	dummy=FLASH_RY_BYBAR;
	
	//select correct bank by driving A16 low
	clrbit(DDRB,5);														//make RB5 output
	FLASH_A16=0;															//set A16 on flash to zero
	reset_flash
}

/*

Erases a complete sector in a 64kword bank of code selected by SelectFlashBank(x)

First set the bank using SelectFlashBank(0..3), this sets address lines a16-a17 to get a block of 64kwords
Next, pass any address in the sector required, and it will erase that complete sector
Note that all addresses must be outside the internal rom of the pic, thus to erase sector 0, you must pass
an address in that sector >=0x1000 for the pic17c43, and >0x2000 for the pic17c44

Example of erasing bank 1,

SelectFlashBank(0);
API_EraseSector(0x2000);

Valid banks (anything else will overwrite program code)

Bank		Absolute Address range			SelectFlashBank(x)					API_EraseSector(x)			Safe to use?
0				0x00000 - 0x01FFF						0														can't do it, inside rom	N/A
1				0x02000 - 0x02FFF						0														0x2000									yes
2				0x03000 - 0x03FFF						0														0x3000									yes
3				0x04000 - 0x07FFF						0														0x4000									yes
4				0x08000 - 0x0FFFF						0														0x8000									yes
5				0x10000 - 0x17FFF						1														0x2000									NO, OVERWRITES PROGRAM CODE
6				0x18000 - 0x1FFFF						1														0x8000									yes
7				0x20000 - 0x27FFF						2														0x2000									NO, OVERWRITES PROGRAM CODE
8				0x28000 - 0x2FFFF						2														0x8000									yes
9				0x30000 - 0x37FFF						3														0x2000									NO, OVERWRITES PROGRAM CODE
10			0x38000 - 0x3FFFF						3														0x8000									yes

*/

void API_EraseSector(unsigned int sector)
{
	void (*bootrom_erase_sector_pointer)(void);									//declare pointer to function

	api_flash_uint_data=sector;																	//pass sector address at top of ram
	bootrom_erase_sector_pointer
		=(void (*)(void))ReadFlash(bootrom_erase_sector_hook);		//hook to vector, see top of this file
	bootrom_erase_sector_pointer();															//equivalent to EraseSector();
}

unsigned char API_WriteFlash(unsigned int addr, unsigned int data)
{
	void (*bootrom_write_flash_pointer)(void);									//declare pointer to function

	api_flash_uint_address=addr;
	api_flash_uint_data=data;
	bootrom_write_flash_pointer
		=(void (*)(void))ReadFlash(bootrom_write_flash_hook);			//hook to vector, see top of this file
	bootrom_write_flash_pointer();															//equivalent to WriteFlash();

	return api_flash_uchar_return;
}

//reads an array of unsigned ints from ram, &ram_addr to rom, rom_addr, of length length
void API_ReadFlashArray(unsigned int rom_addr, unsigned int *ram_addr, unsigned char length)
{
	unsigned char i;

	for(i=0;i<length;i++)
	{
		ram_addr[i]=API_ReadFlash(rom_addr);
		rom_addr++;
	}
}

//writes an array of unsigned ints from ram, &ram_addr to rom, rom_addr, of length length
unsigned char API_WriteFlashArray(unsigned int rom_addr, unsigned int *ram_addr, unsigned char length)
{
	unsigned char i;

	for(i=0;i<length;i++)
	{
		if (API_WriteFlash(rom_addr,ram_addr[i])==FALSE)
		{
			return FALSE;
		}
		rom_addr++;
	}
	return TRUE;
}

//tests routines so far, outputs lots of 'y' for yes if tests pass
void API_TestRoutines(void)
{
	unsigned int flash_array1[4];
	unsigned int flash_array2[4];
	static unsigned int test;

	#define t(x)	if (test==x) {			\
									putch('y');				\
								}										\
								else {							\
									putch('n');				\
									putch(',');				\
									putinthex(test);	\
									putch('!');				\
									putch('=');				\
									putinthex(x);			\
									putch(' ');				\
								}								
									

	API_SetupFlash();
	API_ResetFlash();
	API_SelectFlashBank(0);						//select first bank of 64kwords out of 256kwords

	test=API_FlashDeviceID();
	b;t(0x0002);											//now test=0x0001, give error on purpose to test it
	
	test=API_FlashManufacturerID();	
	b;t(0x22ab);											//now test=0x22ab

	//test for word/byte boundaries
	API_EraseSector(0x2000);
	API_WriteFlash(0x2000,0x0123);
	API_WriteFlash(0x2001,0x3210);
	test=API_ReadFlash(0x2000);					
	b;t(0x0123);											//now test=0x0123
	test=API_ReadFlash(0x2001);					
	b;t(0x3210);											//now test=0x3210

	//test sector 1, 0x2000 to 0x2fff
	API_EraseSector(0x2000);				
	test=API_ReadFlash(0x2000);					
	b;t(0xffff);											//now test=0xFFFF
	API_WriteFlash(0x2000,0x3456);
	test=API_ReadFlash(0x2000);					
	b;t(0x3456);											//now test=0x3456

	//test sector 2, 0x3000 to 0x3fff

	API_EraseSector(0x3000);				
	test=API_ReadFlash(0x3000);					
	b;t(0xffff);											//now test=0xFFFF
	API_WriteFlash(0x3000,0x1234);
	test=API_ReadFlash(0x3000);					
	b;t(0x1234);											//now test=0x1234

	//test sector 4, 0x8000 to 0xffff
	API_EraseSector(0x8000);				
	test=API_ReadFlash(0x8000);					
	b;t(0xffff);											//now test=0xFFFF
	API_WriteFlash(0x8000,0x2345);
	test=API_ReadFlash(0x8000);					
	b;t(0x2345);											//now test=0x2345

	//make sure that data in sector 2 is still intact
	test=API_ReadFlash(0x3000);					
	b;t(0x1234);											//now test=0x1234


	//test write and read array
	#define OFFSET	0x3000
	API_EraseSector(OFFSET);					//0x3000 to 0x3fff
	test=API_ReadFlash(OFFSET);					
	b;t(0xffff);											//now test=0xFFFF

	flash_array1[0]=0x1231;
	flash_array1[1]=0x1232;
	flash_array1[2]=0x1233;
	flash_array1[3]=0x1234;
	b;
	if (API_WriteFlashArray(OFFSET,&flash_array1[0],4)==FALSE)
	{
		b;															//should never stop here
		putch('n');
	}
	b;
	test=API_ReadFlash(OFFSET);
	test=API_ReadFlash(OFFSET+1);
	b;t(0x1232);											//now test=0x1232
	test=API_ReadFlash(OFFSET+2);
	test=API_ReadFlash(OFFSET+3);	
	b;t(0x1234);											//now test=0x1234

	API_ReadFlashArray(OFFSET,&flash_array2[0],4);
	test=flash_array2[0];
	test=flash_array2[1];
	b;t(0x1232);											//now test=0x1232
	test=flash_array2[2];
	test=flash_array2[3];
	b;t(0x1234);											//now test=0x1234
	putch('!');
}

